@@ -98,7 +98,13 @@ where
     }
 
     /// Sends `data` from `peer_addr` to `target_addr`
-    pub async fn send_to(&mut self, peer_addr: SocketAddr, target_addr: Address, data: &[u8]) -> io::Result<()> {
+    #[cfg_attr(not(feature = "local-fake-dns"), allow(unused_mut))]
+    pub async fn send_to(&mut self, peer_addr: SocketAddr, mut target_addr: Address, data: &[u8]) -> io::Result<()> {
+        #[cfg(feature = "local-fake-dns")]
+        if let Some(mapped_addr) = self.context.try_map_fake_address(&target_addr).await {
+            target_addr = mapped_addr;
+        }
+
         // Check or (re)create an association
 
         if let Some(assoc) = self.assoc_map.get(&peer_addr) {
@@ -179,7 +185,7 @@ where
     }
 
     fn try_send(&self, data: (Address, Bytes)) -> io::Result<()> {
-        if let Err(..) = self.sender.try_send(data) {
+        if self.sender.try_send(data).is_err() {
             let err = io::Error::new(ErrorKind::Other, "udp relay channel full");
             return Err(err);
         }
@@ -237,8 +243,9 @@ thread_local! {
     static CLIENT_SESSION_RNG: RefCell<SmallRng> = RefCell::new(SmallRng::from_entropy());
 }
 
+/// Generate an AEAD-2022 Client SessionID
 #[inline]
-fn generate_client_session_id() -> u64 {
+pub fn generate_client_session_id() -> u64 {
     CLIENT_SESSION_RNG.with(|rng| rng.borrow_mut().gen())
 }
 
@@ -376,7 +383,7 @@ where
 
                 _ = keepalive_interval.tick() => {
                     if self.keepalive_flag {
-                        if let Err(..) = self.keepalive_tx.try_send(self.peer_addr) {
+                        if self.keepalive_tx.try_send(self.peer_addr).is_err() {
                             debug!("udp relay {} keep-alive failed, channel full or closed", self.peer_addr);
                         } else {
                             self.keepalive_flag = false;
@@ -473,8 +480,6 @@ where
             target_os = "watchos",
             target_os = "tvos",
             target_os = "freebsd",
-            // target_os = "dragonfly",
-            // target_os = "netbsd",
             target_os = "windows",
         ));
 
@@ -568,8 +573,7 @@ where
                 let svr_cfg = server.server_config();
 
                 let socket =
-                    ProxySocket::connect_with_opts(self.context.context(), svr_cfg, self.context.connect_opts_ref())
-                        .await?;
+                    ProxySocket::connect_with_opts(self.context.context(), svr_cfg, server.connect_opts_ref()).await?;
                 let socket = MonProxySocket::from_socket(socket, self.context.flow_stat());
 
                 self.proxied_socket.insert(socket)
