@@ -4,34 +4,51 @@
 use std::os::unix::io::RawFd;
 use std::{
     io::{self, ErrorKind},
-    net::SocketAddr,
+    net::{IpAddr, SocketAddr},
     sync::Arc,
     time::Duration,
 };
 
 use byte_string::ByteStr;
+use cfg_if::cfg_if;
 use ipnet::IpNet;
 use log::{debug, error, info, trace, warn};
 use shadowsocks::config::Mode;
 use smoltcp::wire::{IpProtocol, TcpPacket, UdpPacket};
-use tokio::{io::AsyncReadExt, sync::mpsc, time};
-use tun::{AsyncDevice, Configuration as TunConfiguration, Device as TunDevice, Error as TunError, Layer};
+use tokio::{
+    io::{AsyncReadExt, AsyncWriteExt},
+    sync::mpsc,
+    time,
+};
+
+cfg_if! {
+    if #[cfg(any(target_os = "ios",
+                 target_os = "macos",
+                 target_os = "linux",
+                 target_os = "android",
+                 target_os = "windows",
+                 target_os = "freebsd"))] {
+        use tun2::{
+            create_as_async, AsyncDevice, Configuration as TunConfiguration, AbstractDevice, Error as TunError, Layer,
+        };
+    } else {
+        use tun2::{AbstractDevice, Configuration as TunConfiguration, Error as TunError, Layer};
+
+        mod fake_tun;
+        use self::fake_tun::{create_as_async, AsyncDevice};
+    }
+}
 
 use crate::local::{context::ServiceContext, loadbalancing::PingBalancer};
 
-use self::{
-    ip_packet::IpPacket,
-    sys::{write_packet_with_pi, IFF_PI_PREFIX_LEN},
-    tcp::TcpTun,
-    udp::UdpTun,
-};
+use self::{ip_packet::IpPacket, tcp::TcpTun, udp::UdpTun};
 
 mod ip_packet;
-mod sys;
 mod tcp;
 mod udp;
 mod virt_device;
 
+/// Tun service builder
 pub struct TunBuilder {
     context: Arc<ServiceContext>,
     balancer: PingBalancer,
@@ -41,7 +58,11 @@ pub struct TunBuilder {
     mode: Mode,
 }
 
+/// TunConfiguration contains a HANDLE, which is a *mut c_void on Windows.
+unsafe impl Send for TunBuilder {}
+
 impl TunBuilder {
+    /// Create a Tun service builder
     pub fn new(context: Arc<ServiceContext>, balancer: PingBalancer) -> TunBuilder {
         TunBuilder {
             context,
@@ -53,52 +74,48 @@ impl TunBuilder {
         }
     }
 
-    pub fn address(mut self, addr: IpNet) -> TunBuilder {
+    pub fn address(&mut self, addr: IpNet) {
         self.tun_config.address(addr.addr()).netmask(addr.netmask());
-        self
     }
 
-    pub fn destination(mut self, addr: IpNet) -> TunBuilder {
+    pub fn destination(&mut self, addr: IpNet) {
         self.tun_config.destination(addr.addr());
-        self
     }
 
-    pub fn name(mut self, name: &str) -> TunBuilder {
-        self.tun_config.name(name);
-        self
+    pub fn name(&mut self, name: &str) {
+        self.tun_config.tun_name(name);
     }
 
     #[cfg(unix)]
-    pub fn file_descriptor(mut self, fd: RawFd) -> TunBuilder {
+    pub fn file_descriptor(&mut self, fd: RawFd) {
         self.tun_config.raw_fd(fd);
-        self
     }
 
-    pub fn udp_expiry_duration(mut self, udp_expiry_duration: Duration) -> TunBuilder {
+    pub fn udp_expiry_duration(&mut self, udp_expiry_duration: Duration) {
         self.udp_expiry_duration = Some(udp_expiry_duration);
-        self
     }
 
-    pub fn udp_capacity(mut self, udp_capacity: usize) -> TunBuilder {
+    pub fn udp_capacity(&mut self, udp_capacity: usize) {
         self.udp_capacity = Some(udp_capacity);
-        self
     }
 
-    pub fn mode(mut self, mode: Mode) -> TunBuilder {
+    pub fn mode(&mut self, mode: Mode) {
         self.mode = mode;
-        self
     }
 
+    /// Build Tun server
     pub async fn build(mut self) -> io::Result<Tun> {
         self.tun_config.layer(Layer::L3).up();
 
-        #[cfg(any(target_os = "linux"))]
-        self.tun_config.platform(|tun_config| {
-            // IFF_NO_PI preventing excessive buffer reallocating
-            tun_config.packet_information(false);
-        });
+        // XXX: tun2 set IFF_NO_PI by default.
+        //
+        // #[cfg(target_os = "linux")]
+        // self.tun_config.platform_config(|tun_config| {
+        //     // IFF_NO_PI preventing excessive buffer reallocating
+        //     tun_config.packet_information(false);
+        // });
 
-        let device = match tun::create_as_async(&self.tun_config) {
+        let device = match create_as_async(&self.tun_config) {
             Ok(d) => d,
             Err(TunError::Io(err)) => return Err(err),
             Err(err) => return Err(io::Error::new(ErrorKind::Other, err)),
@@ -114,7 +131,7 @@ impl TunBuilder {
         let tcp = TcpTun::new(
             self.context,
             self.balancer,
-            device.get_ref().mtu().unwrap_or(1500) as u32,
+            device.as_ref().mtu().unwrap_or(1500) as u32,
         );
 
         Ok(Tun {
@@ -128,6 +145,7 @@ impl TunBuilder {
     }
 }
 
+/// Tun service
 pub struct Tun {
     device: AsyncDevice,
     tcp: TcpTun,
@@ -138,18 +156,52 @@ pub struct Tun {
 }
 
 impl Tun {
+    /// Start serving
     pub async fn run(mut self) -> io::Result<()> {
-        if let Ok(mtu) = self.device.get_ref().mtu() {
-            assert!(mtu > 0 && mtu as usize > IFF_PI_PREFIX_LEN);
-        }
-
         info!(
             "shadowsocks tun device {}, mode {}",
-            self.device.get_ref().name(),
+            self.device
+                .as_ref()
+                .tun_name()
+                .or_else(|r| Ok::<_, ()>(r.to_string()))
+                .unwrap(),
             self.mode,
         );
 
-        let mut packet_buffer = vec![0u8; 65536 + IFF_PI_PREFIX_LEN].into_boxed_slice();
+        let address = match self.device.as_ref().address() {
+            Ok(a) => a,
+            Err(err) => {
+                error!("[TUN] failed to get device address, error: {}", err);
+                return Err(io::Error::new(io::ErrorKind::Other, err));
+            }
+        };
+
+        let netmask = match self.device.as_ref().netmask() {
+            Ok(n) => n,
+            Err(err) => {
+                error!("[TUN] failed to get device netmask, error: {}", err);
+                return Err(io::Error::new(io::ErrorKind::Other, err));
+            }
+        };
+
+        let address_net = match IpNet::with_netmask(address, netmask) {
+            Ok(n) => n,
+            Err(err) => {
+                error!("[TUN] invalid address {}, netmask {}, error: {}", address, netmask, err);
+                return Err(io::Error::new(io::ErrorKind::Other, err));
+            }
+        };
+
+        trace!(
+            "[TUN] tun device network: {} (address: {}, netmask: {})",
+            address_net,
+            address,
+            netmask
+        );
+
+        let address_broadcast = address_net.broadcast();
+
+        let mut packet_buffer = vec![0u8; 65536].into_boxed_slice();
         let mut udp_cleanup_timer = time::interval(self.udp_cleanup_interval);
 
         loop {
@@ -158,28 +210,27 @@ impl Tun {
                 n = self.device.read(&mut packet_buffer) => {
                     let n = n?;
 
-                    if n <= IFF_PI_PREFIX_LEN {
-                        error!(
-                            "[TUN] packet too short, packet: {:?}",
-                            ByteStr::new(&packet_buffer[..n])
-                        );
-                        continue;
-                    }
-
-                    let packet = &mut packet_buffer[IFF_PI_PREFIX_LEN..n];
+                    let packet = &mut packet_buffer[..n];
                     trace!("[TUN] received IP packet {:?}", ByteStr::new(packet));
 
-                    if let Err(err) = self.handle_tun_frame(packet).await {
+                    if let Err(err) = self.handle_tun_frame(&address_broadcast, packet).await {
                         error!("[TUN] handle IP frame failed, error: {}", err);
                     }
                 }
 
                 // UDP channel sent back
                 packet = self.udp.recv_packet() => {
-                    if let Err(err) = write_packet_with_pi(&mut self.device, &packet).await {
-                        error!("[TUN] failed to set packet information, error: {}, {:?}", err, ByteStr::new(&packet));
-                    } else {
-                        trace!("[TUN] sent IP packet (UDP) {:?}", ByteStr::new(&packet));
+                    match self.device.write(&packet).await {
+                        Ok(n) => {
+                            if n < packet.len() {
+                                warn!("[TUN] sent IP packet (UDP), but truncated. sent {} < {}, {:?}", n, packet.len(), ByteStr::new(&packet));
+                            } else {
+                                trace!("[TUN] sent IP packet (UDP) {:?}", ByteStr::new(&packet));
+                            }
+                        }
+                        Err(err) => {
+                            error!("[TUN] failed to set packet information, error: {}, {:?}", err, ByteStr::new(&packet));
+                        }
                     }
                 }
 
@@ -196,17 +247,24 @@ impl Tun {
 
                 // TCP channel sent back
                 packet = self.tcp.recv_packet() => {
-                    if let Err(err) = write_packet_with_pi(&mut self.device, &packet).await {
-                        error!("[TUN] failed to set packet information, error: {}, {:?}", err, ByteStr::new(&packet));
-                    } else {
-                        trace!("[TUN] sent IP packet (TCP) {:?}", ByteStr::new(&packet));
+                    match self.device.write(&packet).await {
+                        Ok(n) => {
+                            if n < packet.len() {
+                                warn!("[TUN] sent IP packet (TCP), but truncated. sent {} < {}, {:?}", n, packet.len(), ByteStr::new(&packet));
+                            } else {
+                                trace!("[TUN] sent IP packet (TCP) {:?}", ByteStr::new(&packet));
+                            }
+                        }
+                        Err(err) => {
+                            error!("[TUN] failed to set packet information, error: {}, {:?}", err, ByteStr::new(&packet));
+                        }
                     }
                 }
             }
         }
     }
 
-    async fn handle_tun_frame(&mut self, frame: &[u8]) -> smoltcp::wire::Result<()> {
+    async fn handle_tun_frame(&mut self, device_broadcast_addr: &IpAddr, frame: &[u8]) -> smoltcp::wire::Result<()> {
         let packet = match IpPacket::new_checked(frame)? {
             Some(packet) => packet,
             None => {
@@ -215,6 +273,32 @@ impl Tun {
             }
         };
 
+        trace!("[TUN] {:?}", packet);
+
+        let src_ip_addr = packet.src_addr();
+        let dst_ip_addr = packet.dst_addr();
+        let src_non_unicast = src_ip_addr == *device_broadcast_addr
+            || match src_ip_addr {
+                IpAddr::V4(v4) => v4.is_broadcast() || v4.is_multicast() || v4.is_unspecified(),
+                IpAddr::V6(v6) => v6.is_multicast() || v6.is_unspecified(),
+            };
+        let dst_non_unicast = dst_ip_addr == *device_broadcast_addr
+            || match dst_ip_addr {
+                IpAddr::V4(v4) => v4.is_broadcast() || v4.is_multicast() || v4.is_unspecified(),
+                IpAddr::V6(v6) => v6.is_multicast() || v6.is_unspecified(),
+            };
+
+        if src_non_unicast || dst_non_unicast {
+            trace!(
+                "[TUN] IP packet {} (unicast? {}) -> {} (unicast? {}) throwing away",
+                src_ip_addr,
+                !src_non_unicast,
+                dst_ip_addr,
+                !dst_non_unicast
+            );
+            return Ok(());
+        }
+
         match packet.protocol() {
             IpProtocol::Tcp => {
                 if !self.mode.enable_tcp() {
@@ -242,7 +326,14 @@ impl Tun {
                 let src_addr = SocketAddr::new(packet.src_addr(), src_port);
                 let dst_addr = SocketAddr::new(packet.dst_addr(), dst_port);
 
-                trace!("[TUN] TCP packet {} -> {} {}", src_addr, dst_addr, tcp_packet);
+                trace!(
+                    "[TUN] TCP packet {} (unicast? {}) -> {} (unicast? {}) {}",
+                    src_addr,
+                    !src_non_unicast,
+                    dst_addr,
+                    !dst_non_unicast,
+                    tcp_packet
+                );
 
                 // TCP first handshake packet.
                 if let Err(err) = self.tcp.handle_packet(src_addr, dst_addr, &tcp_packet).await {
@@ -277,11 +368,18 @@ impl Tun {
                 let src_port = udp_packet.src_port();
                 let dst_port = udp_packet.dst_port();
 
-                let src_addr = SocketAddr::new(packet.src_addr(), src_port);
+                let src_addr = SocketAddr::new(src_ip_addr, src_port);
                 let dst_addr = SocketAddr::new(packet.dst_addr(), dst_port);
 
                 let payload = udp_packet.payload();
-                trace!("[TUN] UDP packet {} -> {} {}", src_addr, dst_addr, udp_packet);
+                trace!(
+                    "[TUN] UDP packet {} (unicast? {}) -> {} (unicast? {}) {}",
+                    src_addr,
+                    !src_non_unicast,
+                    dst_addr,
+                    !dst_non_unicast,
+                    udp_packet
+                );
 
                 if let Err(err) = self.udp.handle_packet(src_addr, dst_addr, payload).await {
                     error!("handle UDP packet failed, err: {}, packet: {:?}", err, udp_packet);
