@@ -98,9 +98,14 @@ impl ProxySocket {
     ) -> ProxySocketResult<ProxySocket> {
         // Note: Plugins doesn't support UDP relay
 
-        let socket = ShadowUdpSocket::connect_server_with_opts(&context, svr_cfg.addr(), opts).await?;
+        let socket = ShadowUdpSocket::connect_server_with_opts(&context, svr_cfg.udp_external_addr(), opts).await?;
 
-        trace!("connected udp remote {} with {:?}", svr_cfg.addr(), opts);
+        trace!(
+            "connected udp remote {} (outbound: {}) with {:?}",
+            svr_cfg.addr(),
+            svr_cfg.udp_external_addr(),
+            opts
+        );
 
         Ok(ProxySocket::from_socket(
             UdpSocketType::Client,
@@ -158,7 +163,7 @@ impl ProxySocket {
         opts: AcceptOpts,
     ) -> ProxySocketResult<ProxySocket> {
         // Plugins doesn't support UDP
-        let socket = match svr_cfg.addr() {
+        let socket = match svr_cfg.udp_external_addr() {
             ServerAddr::SocketAddr(sa) => ShadowUdpSocket::listen_with_opts(sa, opts).await?,
             ServerAddr::DomainName(domain, port) => {
                 lookup_then!(&context, domain, *port, |addr| {
@@ -453,6 +458,7 @@ impl ProxySocket {
     /// This function will use `recv_buf` to store intermediate data, so it has to be big enough to store the whole shadowsocks' packet
     ///
     /// It is recommended to allocate a buffer to have at least 65536 bytes.
+    #[allow(clippy::type_complexity)]
     pub async fn recv_from(&self, recv_buf: &mut [u8]) -> ProxySocketResult<(usize, SocketAddr, Address, usize)> {
         self.recv_from_with_ctrl(recv_buf)
             .await
@@ -464,6 +470,7 @@ impl ProxySocket {
     /// This function will use `recv_buf` to store intermediate data, so it has to be big enough to store the whole shadowsocks' packet
     ///
     /// It is recommended to allocate a buffer to have at least 65536 bytes.
+    #[allow(clippy::type_complexity)]
     pub async fn recv_from_with_ctrl(
         &self,
         recv_buf: &mut [u8],
@@ -497,6 +504,7 @@ impl ProxySocket {
 
     /// poll family functions.
     /// the recv_timeout is ignored.
+    #[allow(clippy::type_complexity)]
     pub fn poll_recv(
         &self,
         cx: &mut Context<'_>,
@@ -507,6 +515,7 @@ impl ProxySocket {
     }
 
     /// poll family functions
+    #[allow(clippy::type_complexity)]
     pub fn poll_recv_with_ctrl(
         &self,
         cx: &mut Context<'_>,
@@ -522,6 +531,8 @@ impl ProxySocket {
         }
     }
 
+    /// poll family functions
+    #[allow(clippy::type_complexity)]
     pub fn poll_recv_from(
         &self,
         cx: &mut Context<'_>,
@@ -531,6 +542,8 @@ impl ProxySocket {
             .map(|r| r.map(|(n, sa, a, rn, _)| (n, sa, a, rn)))
     }
 
+    /// poll family functions
+    #[allow(clippy::type_complexity)]
     pub fn poll_recv_from_with_ctrl(
         &self,
         cx: &mut Context<'_>,
@@ -545,6 +558,7 @@ impl ProxySocket {
         }
     }
 
+    /// poll family functions
     pub fn poll_recv_ready(&self, cx: &mut Context<'_>) -> Poll<ProxySocketResult<()>> {
         self.socket.poll_recv_ready(cx).map_err(|x| x.into())
     }
